import json
import uuid
from datetime import datetime, timedelta

from sqlalchemy import TIMESTAMP, BigInteger, Boolean, Column, String, Text
from sqlalchemy.dialects import mysql, postgresql, sqlite
from sqlalchemy.dialects.postgresql import UUID as psqlUUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.types import BINARY, TypeDecorator
from sqlalchemy.sql import functions as func

Base = declarative_base()

__all__ = ["StagedMessageRepository"]


BigIntegerType = BigInteger()
BigIntegerType = BigIntegerType.with_variant(postgresql.BIGINT(), 'postgresql')
BigIntegerType = BigIntegerType.with_variant(mysql.BIGINT(), 'mysql')
BigIntegerType = BigIntegerType.with_variant(sqlite.INTEGER(), 'sqlite')


class JsonEncodedDict(TypeDecorator):
    """Enables JSON storage by encoding and decoding on the fly."""
    impl = Text

    def process_bind_param(self, value, dialect):
        from . import TelstarEncoder

        if value is None:
            return '{}'
        else:
            return json.dumps(value, cls=TelstarEncoder)

    def process_result_value(self, value, dialect):
        if value is None:
            return {}
        else:
            return json.loads(value)


class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    BINARY(16), to store UUID.

    """
    impl = BINARY

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(psqlUUID())
        else:
            return dialect.type_descriptor(BINARY(16))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                if isinstance(value, bytes):
                    value = uuid.UUID(bytes=value)
                elif isinstance(value, int):
                    value = uuid.UUID(int=value)
                elif isinstance(value, str):
                    value = uuid.UUID(value)
        if dialect.name == 'postgresql':
            return str(value)
        else:
            return value.bytes

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if dialect.name == 'postgresql':
            return uuid.UUID(value)
        else:
            return uuid.UUID(bytes=value)


class StagedMessage(Base):
    __tablename__ = 'telstar_staged_message'

    id = Column(BigIntegerType, primary_key=True)
    msg_uid = Column(UUID(), index=True, nullable=False, default=lambda: uuid.uuid4())
    topic = Column(String(length=255), index=True, nullable=False)
    data = Column(JsonEncodedDict, nullable=False)

    sent = Column(Boolean(), default=False, index=True)
    send_at = Column(TIMESTAMP())
    created_at = Column(TIMESTAMP(), server_default=func.now())

    def to_telstar(self):
        from . import Message
        return Message(self.topic, self.msg_uid, self.data)


class _StagedMessageRepository:
    def __init__(self):
        self.model: StagedMessage = StagedMessage

    def create(self, **kwargs):
        if "delay" in kwargs:
            delay = kwargs.pop("delay")
            kwargs["send_at"] = datetime.now() + timedelta(seconds=delay)
        obj = self.model(**kwargs)
        self.db.add(obj)
        return obj

    def setup(self, database):
        self.db = database

    def get_transaction_wrapper(self):
        return self.db.begin

    def unsent(self):
        # We look into past when sending
        # If you remove the timedelta we get strange errors
        return self.db.query(self.model).filter(self.model.sent == False, self.model.send_at <= datetime.now() + timedelta(seconds=1)).order_by(self.model.id)

    def mark_as_sent(self, messages):
        for m in messages:
            m.sent = True


StagedMessageRepository = _StagedMessageRepository()



import logging
import os
import random
import sys
from time import sleep

import pymysql
import redis
from sqlalchemy import BIGINT, Column, String, UniqueConstraint, create_engine
from sqlalchemy.orm import sessionmaker

from telstar import config as tlconfig
from telstar.com import Message
from telstar.com.sqla import Base, StagedMessageRepository
from telstar.consumer import MultiConsumer

logger = logging.getLogger('telstar')
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)

pymysql.install_as_MySQLdb()

link = redis.from_url(os.environ["REDIS"])

engine = create_engine(os.environ["DATABASE"])

Session = sessionmaker(bind=engine)
session = Session(autocommit=True)

tlconfig.staging.repository = StagedMessageRepository
tlconfig.staging.repository.setup(session)


class Test(Base):
    __tablename__ = "test"
    id = Column(BIGINT(), primary_key=True)
    number = Column(BIGINT)
    group_name = Column(String(length=80))
    topic = Column(String(length=80))

    __table_args__ = (UniqueConstraint('number', 'group_name', 'topic', name='all_together'),)


if __name__ == "__main__":

    if len(sys.argv) > 1 and sys.argv[1] == "setup":
        print("Recreating table in order to start from scratch")
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)

    def simple(consumer, record: Message, done):
        with session.begin():
            t = Test(number=int(record.data["value"]), group_name=consumer.group_name, topic=record.stream)
            session.add(t)
            sleep(random.randrange(int(os.environ.get("SLEEPINESS"))) / 100)
            done()

    MultiConsumer(link=link,
                  group_name=os.environ.get("GROUP_NAME"),
                  consumer_name=os.environ.get("CONSUMER_NAME"),
                  config={
                      os.environ["STREAM_NAME"]: simple,
                      os.environ["STREAM_NAME_TWO"]: simple
                  }).run()



from .com.pw import StagedMessage

__all__ = ["staging"]


class _staging:
    repository = StagedMessage

    def __getattribute__(self, name):
        return _staging.repository

    def __setattr__(self, name, value):
        _staging.repository = value


staging = _staging()



import os
import uuid
import time
from datetime import datetime
from unittest import mock

import peewee
import pytest
import redis
import pymysql
from marshmallow import Schema, ValidationError, fields
from playhouse.db_url import connect
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import StatementError

import telstar
from telstar import config as tlconfig
from telstar.com import Message, MessageError
from telstar.com.pw import StagedMessage as StagedMessagePeeWee
from telstar.com.sqla import StagedMessageRepository as StagedMessageSqlAlchemy
from telstar.consumer import Consumer, MultiConsumeOnce, MultiConsumer
from telstar.producer import StagedProducer

pymysql.install_as_MySQLdb()

def pytest_configure(config):
    config.addinivalue_line(
        "integration", "uses real redis and mysql"
    )


@pytest.fixture
def msg_schema() -> Schema:
    class MyObjSchema(Schema):
        name = fields.Str()
        email = fields.Email()
    return MyObjSchema


@pytest.fixture
def link() -> redis.Redis:
    return mock.MagicMock(spec=redis.Redis)


@pytest.fixture
def reallink() -> redis.Redis:
    client = redis.from_url(os.environ.get("REDIS", "redis://localhost:6379/10"))
    client.flushdb()
    return client


def peewee_db_setup(connection_uri):
    tables = [tlconfig.staging.repository]
    db = connect(connection_uri)
    db.bind(tables)
    db.drop_tables(tables)
    db.create_tables(tables)
    return db


def sqlalchemy_db_setup(connection_uri):
    engine = create_engine(connection_uri)
    return engine

pytest.mark.only_sqla = pytest.mark.skipif(os.environ.get("ORM") != "sqlalchemy", reason="Not supported by peewee")
pytest.mark.only_peewee = pytest.mark.skipif(os.environ.get("ORM") == "sqlalchemy", reason="Not supported by sql")

@pytest.fixture(scope="session")
def db_engine():
    connection_uri = os.environ.get("DATABASE", "postgres://127.0.0.1:5432/telstar-integration-test")
    if os.environ.get("ORM") == "peewee":
        tlconfig.staging.repository = StagedMessagePeeWee
        yield peewee_db_setup(connection_uri)

    if os.environ.get("ORM") == "sqlalchemy":
        from telstar.com.sqla import Base
        tlconfig.staging.repository = StagedMessageSqlAlchemy
        db_engine = sqlalchemy_db_setup(connection_uri)
        Base.metadata.create_all(db_engine)

        yield db_engine

        Base.metadata.drop_all(db_engine)


@pytest.fixture
def session_maker(db_engine) -> peewee.Database:
    if os.environ.get("ORM") == "peewee":
        yield db_engine

    if os.environ.get("ORM") == "sqlalchemy":
        connection = db_engine.connect()
        Session = sessionmaker(bind=db_engine)

        yield Session

        connection.close()


@pytest.fixture
def db_session(session_maker) -> peewee.Database:
    if os.environ.get("ORM") == "peewee":
        with session_maker.transaction() as txn:
            yield session_maker
            txn.rollback()

    if os.environ.get("ORM") == "sqlalchemy":
        session = session_maker()
        tlconfig.staging.repository.setup(session)

        yield session

        session.rollback()
        session.close()


@pytest.fixture
def consumer(link) -> Consumer:
    return Consumer(link, "mygroup", "myname", "mytopic", lambda msg, done: done())


def test_message():
    uid = uuid.uuid4()
    m = Message("topic", uid, dict())
    assert m.msg_uuid == uid


def test_consumer_create_group(link):
    Consumer(link, "mygroup", "myname", "mytopic", lambda msg, done: done())
    link.xgroup_create.assert_called_once_with('telstar:stream:mytopic', 'mygroup', id='0', mkstream=True)


def test_consumer_run(link: redis.Redis):
    callback = mock.Mock()
    msg_id = str(uuid.uuid4()).encode("ascii")
    link.get.return_value = None
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.transfer_and_process_stream_history = lambda *a, **kw: None
    c.run_once()
    callback.assert_called()


def test_consumer_run_callback(link: redis.Redis):
    called = False
    msg_id: bytes = str(uuid.uuid4()).encode("ascii")

    def callback(c, msg: Message, done):
        nonlocal called, msg_id
        called = True
        assert msg.msg_uuid == uuid.UUID(msg_id.decode("ascii"))
        assert msg.data == {}
        assert msg.stream == "mytopic"
        done()

    link.get.return_value = None
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.transfer_and_process_stream_history = lambda *a, **kw: None
    c.run_once()

    assert called is True


def test_consumer_checkpoint(link: redis.Redis):
    msg_id: bytes = b"49ccc393-1594-4d06-b357-0e1f322300b2"

    def callback(c, msg: Message, done):
        done()

    link.get.return_value = None
    pipeline = mock.MagicMock(spec=redis.client.Pipeline)()
    link.pipeline.return_value = pipeline
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.run_once()
    link.get.assert_any_call("telstar:checkpoint:telstar:stream:mytopic:cg:mygroup:myname")
    pipeline.set.assert_called_with("telstar:checkpoint:telstar:stream:mytopic:cg:mygroup:myname", "stream_msg_id")


def test_consumer_with_multiple_stearms(link):
    callback1 = mock.Mock()
    callback2 = mock.Mock()

    config = {
        "mytopic1": callback1,
        "mytopic2": callback2
    }
    msg_id = str(uuid.uuid4()).encode("ascii")
    link.get.return_value = None
    link.xreadgroup.return_value = [
        [b"telstar:stream:mytopic1", [
            ["stream_msg_id1", {b'message_id': msg_id, b"data": "{}"}],
            ["stream_msg_id2", {b'message_id': msg_id, b"data": "{}"}]
        ]],

        [b"telstar:stream:mytopic2", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]]
    ]

    mc = MultiConsumer(link, "group", "name", config)
    mc.transfer_and_process_stream_history = lambda *a, **kw: None
    mc.run_once()

    assert callback1.call_count == 2
    assert callback2.call_count == 1


def test_seen_key(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message("mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_seen_key_bytes(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message(b"mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_message_strip_telstar_prefix(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message(b"telstar:stream:mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_message_with_non_uuid():
    uid = "something random"
    with pytest.raises(TypeError):
        Message("mytopic", uid, dict())


def test_checkpoint_key(consumer: Consumer):
    assert consumer._checkpoint_key("mytopic") == "telstar:checkpoint:mytopic:cg:mygroup:myname"


@pytest.mark.only_peewee
def test_staged_event(db_session):
    telstar.stage("mytopic", dict(a=1))
    assert len(tlconfig.staging.repository.select().where(tlconfig.staging.repository.topic == "mytopic")) == 1


def test_staged_producer(db_session, link):
    telstar.stage("mytopic", dict(a=1))
    [msgs], _ = StagedProducer(link, db_session).get_records()
    assert msgs.stream == "mytopic"
    assert msgs.data == dict(a=1)


def test_encoding_raises_correct_type_error(db_session, link):
    now = datetime.now()

    Error = TypeError
    if os.environ["ORM"] == "sqlalchemy":
        Error = StatementError

    with pytest.raises(Error):
        telstar.stage("mytopic", dict(dt=now, mock=mock.MagicMock()))
        db_session.commit()


def test_stage_can_encode_types(db_session, link):
    now = datetime.now()
    uid = uuid.uuid4()
    telstar.stage("mytopic", dict(dt=now, uuid=uid))
    [msg], cb = StagedProducer(link, db_session).get_records()
    assert msg.data == {"dt": now.isoformat(), "uuid": str(uid)}


def test_staged_producer_done_callback_removes_staged_events(db_session, link):
    telstar.stage("mytopic", dict(a=1))
    telstar.stage("mytopic", dict(a=1))
    msgs, cb = StagedProducer(link, db_session, batch_size=1).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 2
    cb()
    msgs, _ = StagedProducer(link, db_session).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 1


@pytest.mark.only_sqla
def test_staged_producer_context_callable(session_maker, link):
    session = session_maker(autocommit=True)
    tlconfig.staging.repository.setup(session)

    with session.begin():
        telstar.stage("mytopic", dict(a=1))
        telstar.stage("mytopic", dict(b=1))

    sp = StagedProducer(link, session, batch_size=10)

    with sp.context_callable():
        sp.run_once()

    with sp.context_callable():
        sp.run_once()

def test_staged_producer_delay_sending_message(db_session, link):
    telstar.stage("mytopic", dict(a=1), delay=4)
    telstar.stage("mytopic", dict(b=1))
    msgs, cb = StagedProducer(link, db_session, batch_size=10).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 1
    cb()
    msgs, _ = StagedProducer(link, db_session).get_records()
    assert len(msgs) == 0
    assert len(telstar.staged()) == 0
    time.sleep(5)
    msgs, _ = StagedProducer(link, db_session).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 1

def test_consumer_once_keys(link):
    callback = mock.Mock()
    m = MultiConsumeOnce(link, "testgroup", {"mystream": callback})
    assert m._applied_key() == "telstar:once:testgroup"


@pytest.mark.integration
def test_app_pattern(db_session, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", ["mytopic", "mytopic2"], schema=msg_schema)
    def callback(data: dict):
        m()

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    telstar.stage("mytopic2", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, db_session).run_once()

    app.run_once()
    assert m.call_count == 2


@pytest.mark.integration
def test_app_consumer_strictness(db_session, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, db_session).run_once()

    with pytest.raises(ValidationError):
        app.run_once()


@pytest.mark.integration
def test_app_consumer_errorhandler(db_session, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(ValidationError)
    def handler(exc, ack, rec):
        m()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, db_session).run_once()

    app.run_once()
    assert m.call_count == 1


@pytest.mark.integration
def test_app_consumer_errorhandler_can_acknowledge(db_session, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(ValidationError)
    def handler(exc, ack, rec):
        m()
        ack()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, db_session).run_once()

    app.run_once()
    app.run_once()
    assert m.call_count == 1  # The message is ack'ed in the error handler

@pytest.mark.integration
def test_app_consumer_errorhandler_receives_record(db_session, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(ValidationError)
    def handler(exc, ack, record):
        m(record)
        ack()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, db_session).run_once()

    app.run_once()
    assert m.call_count == 1  # The message is ack'ed in the error handler
    [msg, ] = m.call_args[0]
    assert msg[b"data"] == b'{"name": "1", "email": "invalid"}'


@pytest.mark.integration
def test_app_consumer_invalid_message(db_session, reallink: redis.Redis, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(MessageError)
    def handler(exc, ack, rec):
        m()
        ack()

    reallink.xadd("telstar:stream:mytopic", {"not_uuid": "asd", "not_data": "asd"})
    app.run_once()
    app.run_once()
    assert m.call_count == 1  # The message is ack'ed in the error handler


@pytest.mark.integration
def test_app_consumer_ack_invalid(db_session, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=True, strict=False)
    def callback(data: dict):
        pass

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, db_session).run_once()

    ack = mocker.spy(MultiConsumer, "acknowledge")
    app.run_once()
    ack.assert_called_once()


@pytest.mark.integration
def test_app_consumer_do_not_ack_invalid(db_session, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: dict):
        m()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, db_session).run_once()

    ack = mocker.spy(MultiConsumer, "acknowledge")
    app.run_once()
    app.run_once()
    ack.assert_not_called()


@pytest.mark.integration
def test_app_consumer_full_message(db_session, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        m(data)

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, db_session).run_once()

    app.run_once()
    [msg, ] = m.call_args[0]
    assert msg.data == dict(name="1", email="a@b.com")
    assert isinstance(msg, Message)


@pytest.mark.integration
def test_consumer_once(db_session, reallink):
    result = list()
    for i in range(10):
        telstar.stage("mytopic", dict(i=i))

    def callback(c, msg: Message, done):
        data = int(msg.data["i"])
        if data < c.stop:
            result.append(data)
            done()

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()

    m = MultiConsumeOnce(reallink, "mytest", {"mytopic": callback})
    m.stop = 5
    assert m.run() == 10  # Successfully proccessed 10 message but only five got ack'ed

    m.stop = 10
    assert m.run() == 5  # Processes the remaining five
    assert m.run() == 0  # Nothing to process anylonger

    # The message should appear in the order they where sent in.
    assert result == list(range(10))


@pytest.mark.integration
def test_admin_basics(reallink, db_session, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()

    [stream] = admin.get_streams()
    # We have not yet read something from the stream thus there is now group and no consumer
    assert stream.get_pending_messages() == []
    assert stream.get_groups() == []

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        raise Exception("Wont't process")

    with pytest.raises(Exception):
        app.run_once()
    with pytest.raises(Exception):
        app.run_once()

    [msg] = stream.get_pending_messages()
    [grp] = stream.get_groups()
    [consumer] = grp.get_consumers()

    assert stream.get_length() == 1
    assert grp.get_seen_messages() == 0
    assert stream.display_name == b"mytopic"

    assert msg.times_delivered == 2
    assert consumer.idle_time < 1000


@pytest.mark.integration
def test_admin_group_deletion(reallink, db_session, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema)
    def callback(data: Message):
        pass

    app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()

    grp.delete()

    assert stream.get_groups() == []


@pytest.mark.integration
def test_admin_read_pending_message(reallink, db_session, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    data = dict(name="1", email="a@b.com")
    telstar.stage("mytopic", data)

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        raise Exception("Wont't process")

    with pytest.raises(Exception):
        app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()
    [msg] = grp.get_pending_messages()
    assert msg.read().data == data


@pytest.mark.integration
def test_admin_consumer_deletion(reallink, db_session, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema)
    def callback(data: Message):
        pass

    app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()
    [consumer] = grp.get_consumers()

    consumer.delete()

    assert grp.get_consumers() == []


@pytest.mark.integration
def test_consume_order(db_session, reallink):
    result = list()
    telstar.stage("mytopic", dict(i=1))
    telstar.stage("mytopic2", dict(i=2))
    telstar.stage("mytopic2", dict(i=3))
    telstar.stage("mytopic", dict(i=4))
    telstar.stage("mytopic2", dict(i=5))
    telstar.stage("mytopic2", dict(i=6))
    telstar.stage("mytopic2", dict(i=7))
    telstar.stage("mytopic", dict(i=8))
    telstar.stage("mytopic", dict(i=9))
    telstar.stage("mytopic2", dict(i=10))
    telstar.stage("mytopic2", dict(i=11))
    telstar.stage("mytopic2", dict(i=12))

    def callback(c, msg: Message, done):
        data = int(msg.data["i"])
        result.append(data)
        done()

    sp = StagedProducer(reallink, db_session, batch_size=100)
    sp.run_once()
    m = MultiConsumeOnce(reallink, "mytest", {"mytopic": callback, "mytopic2": callback})
    m.run()

    def monotonicity(l):
        # Count the number of times the current element is one smaller than the next
        return sum([x + 1 == y for x, y in zip(l, l[1:])])

    # Maximum monotony
    assert monotonicity(result) >= 3



# from __future__ import annotations
import json
import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Dict, List, Union

import peewee
from peewee import ModelSelect

if TYPE_CHECKING:
    from . import Message


class JSONField(peewee.TextField):
    def db_value(self, value: Dict[str, Union[int, str]]) -> str:
        from . import TelstarEncoder
        return json.dumps(value, cls=TelstarEncoder)

    def python_value(self, value: str) -> Dict[str, Union[int, str]]:
        if value is not None:
            return json.loads(value)


class StagedMessage(peewee.Model):
    msg_uid = peewee.UUIDField(default=uuid.uuid4, index=True)
    topic = peewee.CharField(index=True)
    data = JSONField()

    sent = peewee.BooleanField(default=False, index=True)
    send_at = peewee.TimestampField(resolution=10**3)
    created_at = peewee.TimestampField(resolution=10**3)

    @classmethod
    def create(cls, **kwargs):
        if "delay" in kwargs:
            delay = kwargs.pop("delay")
            kwargs["send_at"] = datetime.now() + timedelta(seconds=delay)
        return super().create(**kwargs)


    @classmethod
    def unsent(cls) -> ModelSelect:
        return cls.select().where(cls.sent == False, cls.send_at <= datetime.now())  # noqa

    @classmethod
    def mark_as_sent(cls, messages: List["Message"]):
        ids = list(map(lambda m: m.id, messages))
        cls.update(sent=True).where(cls.id << ids).execute()

    @classmethod
    def get_transaction_wrapper(cls):
        return cls._meta.database.atomic

    @classmethod
    def setup(cls, database):
        return cls.bind(database)

    def to_telstar(self) -> "Message":
        from . import Message
        return Message(self.topic, self.msg_uid, self.data)



from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, Column, String, BIGINT

from functools import partial
import logging
import os
import random
import sys
import uuid
from time import sleep
from typing import Callable, List, Tuple

import redis
import pymysql

from telstar.com import Message
from telstar.producer import Producer
from telstar import config as tlconfig

from telstar.com.sqla import StagedMessageRepository, Base, UUID, JsonEncodedDict

logger = logging.getLogger('telstar')
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)

pymysql.install_as_MySQLdb()


link = redis.from_url(os.environ["REDIS"])

engine = create_engine(os.environ["DATABASE"])

Session = sessionmaker(bind=engine)
session = Session(autocommit=True)

tlconfig.staging.repository = StagedMessageRepository
tlconfig.staging.repository.setup(session)


class Events(Base):
    __tablename__ = "events"
    id = Column(BIGINT(), primary_key=True)
    msg_uid = Column(UUID(), index=True, nullable=False, default=lambda: uuid.uuid4())
    topic = Column(String(length=255), index=True, nullable=False)
    data = Column(JsonEncodedDict, nullable=False)


if __name__ == "__main__":

    if len(sys.argv) > 1:
        if sys.argv[1] == "setup":
            print("Recreating table in order to start from scratch")
            Base.metadata.drop_all(engine)
            Base.metadata.create_all(engine)

        if "create" in sys.argv[1:]:
            with session.begin():
                for i in range(int(os.environ["RANGE_FROM"]), int(os.environ["RANGE_TO"])):
                    e = Events(topic=os.environ["STREAM_NAME"], data=dict(value=i))
                    session.add(e)

    def puller() -> Tuple[List[Message], Callable[[], None]]:
        qs = session.query(Events).order_by(Events.id).limit(5)
        msgs = [Message(e.topic, e.msg_uid, e.data) for e in qs]

        def done():
            if not os.environ.get("KEEPEVENTS"):
                for r in qs:
                    session.delete(r)
            sleep(random.randrange(int(os.environ.get("SLEEPINESS"))) / 10)
        return msgs, done
    print("starting")
    Producer(link, puller, context_callable=session.begin).run()
